@@ -25,8 +25,10 @@ module Network.Mail.Mime
       -- * Utilities
     , addPart
     , addAttachment
+    , addAttachmentCid
     , addAttachments
     , addAttachmentBS
+    , addAttachmentBSCid
     , addAttachmentsBS
     , renderAddress
     , htmlPart
@@ -54,7 +56,7 @@ import qualified Data.Text.Lazy as LT
 import qualified Data.Text.Lazy.Encoding as LT
 import Data.ByteString.Char8 ()
 import Data.Bits ((.&.), shiftR)
-import Data.Char (isAscii)
+import Data.Char (isAscii, isControl)
 import Data.Word (Word8)
 import qualified Data.ByteString as S
 import Data.Text (Text)
@@ -227,11 +229,15 @@ renderAddress :: Address -> Text
 renderAddress address =
     TE.decodeUtf8 $ toByteString $ showAddress address
 
+-- Only accept characters between 33 and 126, excluding colons. [RFC2822](https://tools.ietf.org/html/rfc2822#section-2.2)
+sanitizeFieldName :: S.ByteString -> S.ByteString
+sanitizeFieldName = S.filter (\w -> w >= 33 && w <= 126 && w /= 58)
+
 showHeader :: (S.ByteString, Text) -> Builder
 showHeader (k, v) = mconcat
-    [ fromByteString k
+    [ fromByteString (sanitizeFieldName k)
     , fromByteString ": "
-    , encodeIfNeeded v
+    , encodeIfNeeded (sanitizeHeader v)
     , fromByteString "\n"
     ]
 
@@ -251,12 +257,16 @@ showAddressHeader (k, as) =
 -- Since 0.4.3
 showAddress :: Address -> Builder
 showAddress a = mconcat
-    [ maybe mempty ((`mappend` fromByteString " ") . encodedWord) (addressName a)
+    [ maybe mempty ((<> fromByteString " ") . encodedWord) (addressName a)
     , fromByteString "<"
-    , fromText (addressEmail a)
+    , fromText (sanitizeHeader $ addressEmail a)
     , fromByteString ">"
     ]
 
+-- Filter out control characters to prevent CRLF injection.
+sanitizeHeader :: Text -> Text
+sanitizeHeader = T.filter (not . isControl)
+
 showBoundPart :: Boundary -> (Headers, Builder) -> Builder
 showBoundPart (Boundary b) (headers, content) = mconcat
     [ fromByteString "--"
@@ -430,10 +440,26 @@ htmlPart body = Part cType QuotedPrintableText Nothing [] $ LT.encodeUtf8 body
 -- | Add an attachment from a file and construct a 'Part'.
 addAttachment :: Text -> FilePath -> Mail -> IO Mail
 addAttachment ct fn mail = do
-    content <- L.readFile fn
-    let part = Part ct Base64 (Just $ T.pack (takeFileName fn)) [] content
+    part <- getAttachmentPart ct fn
     return $ addPart [part] mail
 
+-- | Add an attachment from a file and construct a 'Part'
+-- with the specified content id in the Content-ID header.
+--
+-- @since 0.4.12
+addAttachmentCid :: Text -- ^ content type
+                 -> FilePath -- ^ file name
+                 -> Text -- ^ content ID
+                 -> Mail
+                 -> IO Mail
+addAttachmentCid ct fn cid mail =
+  getAttachmentPart ct fn >>= (return.addToMail.addHeader) 
+  where 
+    addToMail part = addPart [part] mail
+    addHeader part = part { partHeaders = header:ph }
+      where ph = partHeaders part
+    header = ("Content-ID", T.concat ["<", cid, ">"])
+
 addAttachments :: [(Text, FilePath)] -> Mail -> IO Mail
 addAttachments xs mail = foldM fun mail xs
   where fun m (c, f) = addAttachment c f m
@@ -446,15 +472,40 @@ addAttachmentBS :: Text -- ^ content type
                 -> L.ByteString -- ^ content
                 -> Mail -> Mail
 addAttachmentBS ct fn content mail =
-    let part = Part ct Base64 (Just fn) [] content
+    let part = getAttachmentPartBS ct fn content
     in addPart [part] mail
 
+-- | @since 0.4.12
+addAttachmentBSCid :: Text -- ^ content type
+                -> Text -- ^ file name
+                -> L.ByteString -- ^ content
+                -> Text -- ^ content ID
+                -> Mail -> Mail
+addAttachmentBSCid ct fn content cid mail =
+    let part = addHeader $ getAttachmentPartBS ct fn content
+    in addPart [part] mail
+    where
+      addHeader part = part { partHeaders = header:ph }
+        where ph = partHeaders part
+      header = ("Content-ID", T.concat ["<", cid, ">"])
+
 -- |
 -- Since 0.4.7
 addAttachmentsBS :: [(Text, Text, L.ByteString)] -> Mail -> Mail
 addAttachmentsBS xs mail = foldl fun mail xs
   where fun m (ct, fn, content) = addAttachmentBS ct fn content m
 
+getAttachmentPartBS :: Text
+                    -> Text
+                    -> L.ByteString
+                    -> Part
+getAttachmentPartBS ct fn content = Part ct Base64 (Just fn) [] content
+
+getAttachmentPart :: Text -> FilePath -> IO Part
+getAttachmentPart ct fn = do
+    content <- L.readFile fn
+    return $ getAttachmentPartBS ct (T.pack (takeFileName fn)) content
+
 data QP = QPPlain S.ByteString
         | QPNewline
         | QPTab
